require "http"
require "bigdecimal"

class RpcError < StandardError
end

# A simple client for sending data to the RPC server.
# This is a low-level API; use NanoAccount instead.
class NanoRpcClient
    # Create a new client.
    # Params:
    # +endpoint+:: The location of the RPC server.
    def initialize(endpoint)
        @dispatcher = HTTP 
        @endpoint = endpoint
    end

    # Make a request to the endpoint.
    # Params:
    # +data+:: A +Hash+ to send to the RPC server.
    # Returns:
    # The object returned by the server as a +Hash+.
    def req(data)
        res = @dispatcher.post(@endpoint, :json => data).parse

        raise RpcError, res["error"] if res["error"] != nil

        res
    end

    # Missing methods will make requests to the RPC server using the method name as the action.
    # The rest of the keyword arguments are passed as JSON data.
    # Params:
    # +action+:: The +action+ field of the RPC request.
    # +kwargs+:: The keyword arguments passed to the method.
    def method_missing(action, **kwargs)
        kwargs["action"] = action.to_s

        # Stringify all keys
        req(kwargs.map{|k, v| [k.to_s, v]}.to_h)
    end

    # Convert an amount of MRai to a raw Nano quantity.
    # Params:
    # +mrai+:: The quantity of MRai.
    # Returns:
    # The quantity of raw Nano, stringified.
    def self.mrai_to_raw(mrai)
        BigDecimal(mrai) * BigDecimal("1000000000000000000000000000000")
    end
end

# A variant of +NanoRpcClient+ which connects public to Nanode servers using an API key.
class NanodeClient < NanoRpcClient
    # Create a new Nanode client.
    # Params: 
    # +key+:: A valid Nanode API key.
    def initialize(key)
        @dispatcher = HTTP.auth(key)
        @endpoint = "https://api.nanode.co/"
    end
end